==> leafnode-1.9.26.rel.tar.bz2: leafnode-1.9.26.rel/nntputil.c <==
/*
nntputil.c -- misc nntp-related stuff

Written by Arnt Gulbrandsen <agulbra@troll.no> and copyright 1995
Troll Tech AS, Postboks 6133 Etterstad, 0602 Oslo, Norway, fax +47
22646949.
Modified by Cornelius Krasel <krasel@wpxx02.toxi.uni-wuerzburg.de>
and Randolf Skerka <Randolf.Skerka@gmx.de>.
Copyright of the modifications 1997.
Modified by Kent Robotti <robotti@erols.com>. Copyright of the
modifications 1998.
Modified by Markus Enzenberger <enz@cip.physik.uni-muenchen.de>.
Copyright of the modifications 1998.
Modified by Cornelius Krasel <krasel@wpxx02.toxi.uni-wuerzburg.de>.
Copyright of the modifications 1998, 1999.
Modified by Matthias Andree <matthias.andree@web.de>.
Copyright of the modifications 2000 - 2002.
Modified by Ralf Wildenhues <ralf.wildenhues@gmx.de>.
Copyright of the modifications 2002.

See file COPYING for restrictions on the use of this software.
*/

#include "leafnode.h"
#include <fcntl.h>
#include <sys/uio.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <netinet/in.h>
#ifndef __LCLINT__
#include <arpa/inet.h>
#endif				/* not __LCLINT__ */
#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <stdlib.h>
#include <netdb.h>
#include <setjmp.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include <dirent.h>

extern struct server *current_server;

char last_command[SIZE_lineout + 1];
char lineout[SIZE_lineout + 1];

/*@relnull@*//*@dependent@*/ FILE *nntpin = NULL;
/*@relnull@*//*@dependent@*/ FILE *nntpout = NULL;

int authenticated;
int stat_is_evil = 0;

extern jmp_buf timeout;

extern RETSIGTYPE timer(int sig);


/*
05/26/97 - T. Sweeney - Send a string out, keeping a copy in reserve.
*/
void
putaline(void)
{
    if (debug) {
	char y, *x = lineout + strcspn(lineout, "\r\n");

	y = *x; *x = '\0';
	syslog(LOG_DEBUG, ">%s", lineout);
	*x = y;
    }
    strcpy(last_command, lineout);	/* RATS: ignore */
    fputs(lineout, nntpout);
    fflush(nntpout);
}

/*
 * Authenticate ourselves at a remote server.
 * Returns TRUE if authentication succeeds, FALSE if it does not.
 */
int
authenticate(void)
{
    int d, reply;

    if (!current_server) {
	syslog(LOG_INFO, "authenticate: unknown server");
	return FALSE;
    }
    if (!current_server->username) {
	syslog(LOG_INFO, "%s: username needed for authentication",
	       current_server->name);
	return FALSE;
    }

    d = debug;
    debug = debugmode;
    fprintf(nntpout, "authinfo user %s\r\n", current_server->username);
    fflush(nntpout);

    reply = nntpreply();
    debug = d;
    if (reply == 281) {
	return TRUE;
    } else if (reply != 381) {
	syslog(LOG_INFO, "username rejected: %03d", reply);
	return FALSE;
    }

    if (!current_server->password) {
	syslog(LOG_INFO, "%s: password needed for authentication",
	       current_server->name);
	return FALSE;
    }
    debug = debugmode;
    fprintf(nntpout, "authinfo pass %s\r\n", current_server->password);
    fflush(nntpout);

    reply = nntpreply();
    debug = d;

    if (reply != 281) {
	syslog(LOG_INFO, "password failed: %03d", reply);
	return FALSE;
    }
    return TRUE;
}

static size_t lllen = 0;
static
				    /*@null@*/
 /*@owned@*/
char *llstr = NULL;

/*@dependent@*//*@null@*/ char *
lastreply(void)
{
    return llstr;
}

/*
 * decode an NNTP reply number
 * reads a line from the server and returns an integer
 *
 * 498 is used to mean "protocol error", like smail
 *
 * the text returned is discarded
 *
 * from Tim Sweeney: retry in case of authinfo failure.
 */

int
nntpreply(void)
{
    char *response;
    int r = 0;
    int c = 1;

    while (c) {
	response = mgetaline(nntpin);
	if (!response) {
	    if (llstr)
		free(llstr);
	    llstr = NULL;
	    syslog(LOG_ERR, "NNTP server went away");
	    return 498;
	}
	if (debug == 1) syslog(LOG_DEBUG, "<%s", response);
	if (strlen(response) > lllen || !llstr) {
	    if (llstr)
		free(llstr);
	    lllen = strlen(response);
	    llstr = critmalloc(lllen + 1, "nntpreply");
	}
	strcpy(llstr, response);	/* RATS: ignore */
	if (strlen(response) > 2 && isdigit((unsigned char)response[0])
	    && isdigit((unsigned char)response[1])
	    && isdigit((unsigned char)response[2])
	    && ((response[3] == ' ')
		|| (response[3] == '\0')
		|| (response[3] == '-'))) {
	    int rl;
	    rl = atoi(response);
	    if (r > 0 && r != rl)
		r = 498;	/* protocol error */
	    else
		r = rl;
	    c = (response[3] == '-');
	} else {
	    c = 0;
	    r = 498;		/* protocol error */
	}
    }

    if (r == 480 && !authenticated) {	/* need to authenticate */
	authenticated = TRUE;
	if (authenticate()) {
	    fprintf(nntpout, "%s", last_command);
	    fflush(nntpout);
	    r = nntpreply();
	}
    }
    return r;
}


extern struct state _res;

#define incopy(a)       (*((struct in_addr *)a))

/*
 * connect to upstream nntp server
 *
 * returns 200 for posting allowed, 201 for read-only;
 * if connection failed, return 0
 */
int
nntpconnect(const struct server *upstream)
{
    struct hostent *hp;
    static /*@observer@*/ struct servent *sp;
    struct servent sp_def;
    struct sockaddr_in s_in;
    int sock, reply;
    static int i;

    memset((void *)&s_in, 0, sizeof(s_in));
    if (upstream->port == 0) {
	sp = getservbyname("nntp", "tcp");
	if (sp == NULL) {
	    syslog(LOG_ERR, "unable to find service NNTP");
	    return FALSE;
	}
    } else {
	sp = &sp_def;
	sp->s_port = htons(upstream->port);
    }

    /* Fetch the ip addresses of the given host. */
    hp = gethostbyname(upstream->name);
    if (hp) {
	char *line;

	/* Try to make connection to each of the addresses in turn. */
	for (i = 0; (int *)(hp->h_addr_list)[i]; i++) {
	    s_in.sin_family = hp->h_addrtype;
	    s_in.sin_port = sp->s_port;
	    s_in.sin_addr = incopy(hp->h_addr_list[i]);

	    sock = socket(AF_INET, SOCK_STREAM, 0);
	    if (sock < 0)
		break;

	    if (setjmp(timeout) != 0) {
		(void)close(sock);
		continue;
	    }

	    (void)signal(SIGALRM, timer);
	    (void)alarm((unsigned)upstream->timeout);
	    if (connect(sock, (struct sockaddr *)&s_in, sizeof(s_in)) < 0)
		break;
	    (void)alarm((unsigned)0);
	    (void)signal(SIGALRM, SIG_DFL);

	    nntpout = fdopen(sock, "w");
	    if (nntpout == NULL)
		break;

	    nntpin = fdopen(dup(sock), "r");
	    if (nntpin == NULL)
		break;

	    reply = nntpreply();
	    if (reply == 200 || reply == 201) {
		syslog(LOG_INFO, "connected to %s:%d, reply: %d",
		       inet_ntoa(s_in.sin_addr), (int)ntohs(s_in.sin_port),
		       reply);
		line = lastreply();
		if (line && strstr(line, "NNTPcache server V2.3")) {
		    /* NNTPcache 2.3.3 is still in widespread use, but it
		     * has Y2k bugs which have only been fixed in a beta
		     * version as of 2001-12-24. This 2.3 version is
		     * unsuitable for any use since 2000-01-01. */
		    static const char msg[] =
			"%s: Server greeting \"%s\" hints to "
			"NNTPcache v2.3.x. "
			"This server has severe bugs which make it "
			"unsuitable for use with leafnode. "
			"Ask the news server administrator to update to "
			"NNTPcache v3.0.x or newer.";
		    syslog(LOG_ERR, msg, upstream->name, line); /* RATS: ignore */
		    fprintf(stderr, msg, upstream->name, line); /* RATS: ignore */
		    nntpdisconnect();
		    return 0;
		}
                /*
		 * NewsCache 0.99.17 and previous versions always
		 * reply with 223 0 <MID> when asked "STAT <MID>". This
		 * is a violation of RFC 977 and breaks posting.
		 */
		if (strstr(lastreply(), "NewsCache"))
		    stat_is_evil = 1;
		else
		    stat_is_evil = 0;
		return reply;
	    }
	    shutdown(fileno(nntpout), 0);
	}			/* end of IP-addresses for loop */
    }
    return FALSE;
}				/* end of connect function */

/*
 * disconnect from upstream server
 */
void
nntpdisconnect(void)
{
    if (nntpin) {
	fclose(nntpin);
	nntpin = NULL;
    }
    if (nntpout) {
	fclose(nntpout);
	nntpout = NULL;
    }
}

==> leafnode-1.9.46.rel.tar.bz2: leafnode-1.9.46.rel/nntputil.c <==
/*
nntputil.c -- misc nntp-related stuff

Written by Arnt Gulbrandsen <agulbra@troll.no> and copyright 1995
Troll Tech AS, Postboks 6133 Etterstad, 0602 Oslo, Norway, fax +47
22646949.
Modified by Cornelius Krasel <krasel@wpxx02.toxi.uni-wuerzburg.de>
and Randolf Skerka <Randolf.Skerka@gmx.de>.
Copyright of the modifications 1997.
Modified by Kent Robotti <robotti@erols.com>. Copyright of the
modifications 1998.
Modified by Markus Enzenberger <enz@cip.physik.uni-muenchen.de>.
Copyright of the modifications 1998.
Modified by Cornelius Krasel <krasel@wpxx02.toxi.uni-wuerzburg.de>.
Copyright of the modifications 1998, 1999.
Modified by Ralf Wildenhues <ralf.wildenhues@gmx.de>.
Copyright of the modifications 2002.
Modified by Matthias Andree <matthias.andree@web.de>.
Copyright of the modifications 2000 - 2003.

See file COPYING for restrictions on the use of this software.
*/

#include "system.h"
#include "leafnode.h"
#include "mysigact.h"

#include <fcntl.h>
#include <sys/uio.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <netinet/in.h>
#ifndef __LCLINT__
#include <arpa/inet.h>
#endif				/* not __LCLINT__ */
#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <stdlib.h>
#include <netdb.h>
#include <setjmp.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <dirent.h>

char last_command[SIZE_lineout + 1];
char lineout[SIZE_lineout + 1];

/*@relnull@*//*@dependent@*/ FILE *nntpin = NULL;
/*@relnull@*//*@dependent@*/ FILE *nntpout = NULL;

int authenticated;
int stat_is_evil = 0;
int date_is_evil = 0;

extern sigjmp_buf timeout;

extern RETSIGTYPE timer(int sig);


/*
05/26/97 - T. Sweeney - Send a string out, keeping a copy in reserve.
*/
void
putaline(void)
{
    if (debug > 1) {
	char y, *x = lineout + strcspn(lineout, "\r\n");

	y = *x; *x = '\0';
	syslog(LOG_DEBUG, ">%s", lineout);
	*x = y;
    }
    strcpy(last_command, lineout);	/* RATS: ignore */
    fputs(lineout, nntpout);
    fflush(nntpout);
}

/*
 * Authenticate ourselves at a remote server.
 * Returns TRUE if authentication succeeds, FALSE if it does not.
 */
int
authenticate(const struct server *current_server)
{
    int d, reply;

    if (!current_server) {
	syslog(LOG_INFO, "authenticate: unknown server");
	return FALSE;
    }
    if (!current_server->username) {
	syslog(LOG_INFO, "%s: username needed for authentication",
	       current_server->name);
	return FALSE;
    }

    d = debug;
    debug = debugmode;
    fprintf(nntpout, "authinfo user %s\r\n", current_server->username);
    fflush(nntpout);

    reply = nntpreply(current_server);
    debug = d;
    if (reply == 281) {
	return TRUE;
    } else if (reply != 381) {
	syslog(LOG_INFO, "username rejected: %03d", reply);
	return FALSE;
    }

    if (!current_server->password) {
	syslog(LOG_INFO, "%s: password needed for authentication",
	       current_server->name);
	return FALSE;
    }
    debug = debugmode;
    fprintf(nntpout, "authinfo pass %s\r\n", current_server->password);
    fflush(nntpout);

    reply = nntpreply(current_server);
    debug = d;

    if (reply != 281) {
	syslog(LOG_INFO, "password failed: %03d", reply);
	return FALSE;
    }
    return TRUE;
}

static size_t lllen = 0;
static
				    /*@null@*/
 /*@owned@*/
char *llstr = NULL;

/*@dependent@*//*@null@*/ char *
lastreply(void)
{
    return llstr;
}

/*
 * decode an NNTP reply number
 * reads a line from the server and returns an integer
 *
 * 498 is used to mean "protocol error", like smail
 *
 * the text returned is discarded
 *
 * from Tim Sweeney: retry in case of authinfo failure.
 */

int
nntpreply(const struct server *current_server)
{
    char *response;
    int r = 0;
    int c = 1;

    while (c) {
	response = mgetaline(nntpin);
	if (!response) {
	    if (llstr)
		free(llstr);
	    llstr = NULL;
	    syslog(LOG_ERR, "NNTP server went away");
	    return 498;
	}
	if (debug == 1) syslog(LOG_DEBUG, "<%s", response);
	if (strlen(response) > lllen || !llstr) {
	    if (llstr)
		free(llstr);
	    lllen = strlen(response);
	    llstr = critmalloc(lllen + 1, "nntpreply");
	}
	strcpy(llstr, response);	/* RATS: ignore */
	if (strlen(response) > 2 && isdigit((unsigned char)response[0])
	    && isdigit((unsigned char)response[1])
	    && isdigit((unsigned char)response[2])
	    && ((response[3] == ' ')
		|| (response[3] == '\0')
		|| (response[3] == '-'))) {
	    int rl;
	    rl = atoi(response);
	    if (r > 0 && r != rl)
		r = 498;	/* protocol error */
	    else
		r = rl;
	    c = (response[3] == '-');
	} else {
	    c = 0;
	    r = 498;		/* protocol error */
	}
    }

    if (r == 480 && !authenticated) {	/* need to authenticate */
	authenticated = TRUE;
	if (authenticate(current_server)) {
	    fprintf(nntpout, "%s", last_command);
	    fflush(nntpout);
	    r = nntpreply(current_server);
	}
    }
    return r;
}

struct versions {
    const char *name;
    int is_evil;
};

/*
 * NewsCache 0.99.17 and previous versions always
 * reply with 223 0 <MID> when asked "STAT <MID>". This
 * is a violation of RFC 977 and breaks posting.
 */
static struct versions stat_versions[] = {
    { "NewsCache 0.99.2 ", 1 },
    { "NewsCache 0.99.2", 0 },
    { "NewsCache 0.99.19", 0 },
    { "NewsCache 0.99.18", 0 },
    { "NewsCache 1.0.", 1 },
    { "NewsCache 1", 0 },
    { "NewsCache", 1 },
    /* reported to be necessary by 
     * Robert Marshall <robert@chezmarshall.freeserve.co.uk>:
     * nc news.cache.ntlworld.com 119
     * 200 ntl NNTP news cache. posting ok (feedback to nntptrial-feedback@ntli.net)
     * quit
     * 205 NNTP Service closing connection - goodbye!
     */
    { "NNTP news cache", 1 },
};
static const int stat_count = sizeof(stat_versions)/sizeof(stat_versions[0]);

static struct versions date_versions[] = {
    { "NewsCache 0.99.22p", 0 },
    { "NewsCache 0.99.2 ", 1 },
    { "NewsCache 0.99.20", 1 },
    { "NewsCache 0.99.21", 1 },
    { "NewsCache 0.99.22", 1 },
    { "NewsCache 0.99.2", 0 },
    { "NewsCache 1.1.10 ", 1 },
    { "NewsCache 1.1.11 ", 1 },
    { "NewsCache 1.1.1 ", 1 },
    { "NewsCache 1.1.0", 1},
    { "NewsCache 1", 0},
    { "NewsCache", 1 }
};
static const int date_count = sizeof(date_versions)/sizeof(date_versions[0]);

static int check_linlist(const char *s, const struct versions *list, int count) {
    int i;

    for (i = 0; i < count; i++) {
	if (strstr(s, list[i].name)) {
	    return list[i].is_evil;
	}
    }
    return 0;
}

extern struct state _res;

#define incopy(a)       (*((struct in_addr *)a))

/*
 * connect to upstream nntp server
 *
 * returns 200 for posting allowed, 201 for read-only;
 * if connection failed, return 0
 */
int
nntpconnect(const struct server *upstream)
{
    struct hostent *hp;
    static /*@observer@*/ struct servent *sp;
    struct servent sp_def;
    struct sockaddr_in s_in;
    int sock, reply;
    static int i;

    memset((void *)&s_in, 0, sizeof(s_in));
    if (upstream->port == 0) {
	sp = getservbyname("nntp", "tcp");
	if (sp == NULL) {
	    syslog(LOG_ERR, "unable to find service NNTP");
	    return FALSE;
	}
    } else {
	sp = &sp_def;
	sp->s_port = htons(upstream->port);
    }

    /* Fetch the ip addresses of the given host. */
    hp = gethostbyname(upstream->name);
    if (hp) {
	char *line;

	/* Try to make connection to each of the addresses in turn. */
	for (i = 0; (int *)(hp->h_addr_list)[i]; i++) {
	    s_in.sin_family = hp->h_addrtype;
	    s_in.sin_port = sp->s_port;
	    s_in.sin_addr = incopy(hp->h_addr_list[i]);

	    sock = socket(AF_INET, SOCK_STREAM, 0);
	    if (sock < 0)
		break;

	    if (sigsetjmp(timeout,1) != 0) {
		(void)close(sock);
		continue;
	    }

	    (void)mysigact(SIGALRM, SA_RESETHAND, timer, 0);
	    (void)alarm((unsigned)upstream->timeout);
	    if (connect(sock, (struct sockaddr *)&s_in, sizeof(s_in)) < 0)
		break;
	    (void)alarm((unsigned)0);
	    (void)mysigact(SIGALRM, 0, SIG_DFL, 0);

	    nntpout = fdopen(sock, "w");
	    if (nntpout == NULL)
		break;

	    nntpin = fdopen(dup(sock), "r");
	    if (nntpin == NULL)
		break;

	    reply = nntpreply(upstream);
	    if (reply == 200 || reply == 201) {
		syslog(LOG_INFO, "connected to %s:%d, reply: %d",
		       inet_ntoa(s_in.sin_addr), (int)ntohs(s_in.sin_port),
		       reply);
		line = lastreply();
		if (line == NULL) {
		    static const char msg[] = "server %s disconnected";
		    syslog(LOG_ERR, msg, upstream->name);
		    fprintf(stderr, msg, upstream->name);
		    fprintf(stderr, "\n");
		    return 0;
		}

		if (strstr(line, "NNTPcache server V2.3")) {
		    /* NNTPcache 2.3.3 is still in widespread use, but it
		     * has Y2k bugs which have only been fixed in a beta
		     * version as of 2001-12-24. This 2.3 version is
		     * unsuitable for any use since 2000-01-01. */
		    static const char msg[] =
			"%s: Server greeting \"%s\" hints to "
			"NNTPcache v2.3.x. "
			"This server has severe bugs which make it "
			"unsuitable for use with leafnode. "
			"Ask the news server administrator to update to "
			"NNTPcache v3.0.x or newer.";
		    syslog(LOG_ERR, msg, upstream->name, line); /* RATS: ignore */
		    fprintf(stderr, msg, upstream->name, line); /* RATS: ignore */
		    fprintf(stderr, "\n");
		    nntpdisconnect();
		    return 0;
		}
		stat_is_evil = check_linlist(lastreply(), stat_versions,
			stat_count);
		date_is_evil = check_linlist(lastreply(), date_versions,
			date_count);
		if (stat_is_evil) {
		    syslog(LOG_WARNING, "warning: server \"%s\" greeting "
			    "\"%s\" hints to "
			    "an outdated version with broken "
			    "STAT command handling. Please ask the upstream "
			    "maintainer to update. Emulating STAT with HEAD.",
			    upstream->name, line);
		}
		return reply;
	    }
	    shutdown(fileno(nntpout), 0);
	}			/* end of IP-addresses for loop */
    }
    return FALSE;
}				/* end of connect function */

/*
 * disconnect from upstream server
 */
void
nntpdisconnect(void)
{
    if (nntpin) {
	fclose(nntpin);
	nntpin = NULL;
    }
    if (nntpout) {
	fclose(nntpout);
	nntpout = NULL;
    }
}

#ifdef MAIN
int verbose=0;
int debug=0;
struct server *current_server = NULL;

int main(int argc, char **argv) {
    int i = 1;
    while (i < argc) {
	int stat_is_evil;
	int date_is_evil;

	stat_is_evil = check_linlist(argv[i], stat_versions,
		stat_count);
	date_is_evil = check_linlist(argv[i], date_versions,
		date_count);

	printf("%s: stat_evil=%d, date_evil=%d\n",
		argv[i], stat_is_evil, date_is_evil);

	i++;
    }
    return 0;
}
#endif
